/*   HANDLING PARTICLES (stars in a "galaxy").
*/

#include <stdlib.h> 
#include <string.h> 
#include <stdio.h>
#include <math.h>
#ifdef _OPENMP          
#include <omp.h>
#endif
#include "common.h"
#include "hmi.h"
#include "init.h"
#include "particles.h"
#include "fft.h"
#include "stet.h"
#include "bulletcluster.h"
#include "potential.h"

/* --------------------------------- Engine variables ------------------------------- */
#if INIT_TYPE != INIT_BULLET_CLUSTER
  particle_t Part[NBMAX_PART];
#endif
double Gravit_dt;                         // Here for debug.
unsigned int NbPart;          // Now calculated dynamically because of complicated init.
unsigned int NbPartStetGas;      // Gas particles.
unsigned int NbPartOut;
double PartMass;
double total_time;
double PosOrigin[3];

#if INIT_TYPE != INIT_BULLET_CLUSTER

#if NB_DIM == 2 && INIT_TYPE
  unsigned int CellNbPart[NBCELLX_GRID+2][NBCELLY_GRID+2][NBCELLZ_GRID];
  unsigned int CellNbPartStetGas[NBCELLX_GRID+2][NBCELLY_GRID+2][NBCELLZ_GRID];
           // +2 for allowing the test of x, y, z out of bounds of regular grid.
#elif NB_DIM == 3
  unsigned int CellNbPart[NBCELLX_GRID+2][NBCELLY_GRID+2][NBCELLZ_GRID+2];
  unsigned int CellNbPartStetGas[NBCELLX_GRID+2][NBCELLY_GRID+2][NBCELLZ_GRID+2];
#endif

/* ----------------------- Prototypes of internal functions -------------------------- */
void GetAccelFromGalaxy(int PartInd, double accel[], unsigned int NumCell[]);

/* ----------------------------- Init of the grid ------------------------------------ */
void InitParticles()
{
  NbPartOut = 0;
  memset(CellNbPart, 0, sizeof(CellNbPart));
  memset(CellNbPartStetGas, 0, sizeof(CellNbPartStetGas));
  VectorZero(PosOrigin);
}

/* --------------------- Get the cell containing a given particle ----------- */
unsigned char GetCellNonPeriod(double pos[], unsigned int NumCell[])
{
  double CellOffset[3];
  unsigned char ReturnCode;

  /* --------------------- Calculate NumCell if IN ------------------------- */
  CellOffset[0] = pos[0] - GRID_ORIGIN0 - PosOrigin[0];    
  if(CellOffset[0] < 0  ||  CellOffset[0] > LG_GRID)
  {
    ReturnCode = CELL_OUTX;
  }
  else
  {
    CellOffset[1] = pos[1] - GRID_ORIGIN1 - PosOrigin[1];    
    if(CellOffset[1] < 0  ||  CellOffset[1] > LG_GRID)
    {
      ReturnCode = CELL_OUTY;
    }
    else
    {
      NumCell[0] = (unsigned int) (CellOffset[0] / LGX_CELL);
      NumCell[1] = (unsigned int) (CellOffset[1] / LGY_CELL);
      #if (NB_DIM == 3)
        CellOffset[2] = pos[2] - GRID_ORIGIN2 - PosOrigin[2];    
        if(CellOffset[2] < 0  ||  CellOffset[2] > LG_GRID)
        {
          ReturnCode = CELL_OUTZ;
        }
        else
        {
          NumCell[2] = (unsigned int) (CellOffset[2] / LGZ_CELL);
      #else
          NumCell[2] = 0;                 
      #endif
          if(GridDisk[NumCell[0]][NumCell[1]][NumCell[2]])
          {
            ReturnCode = CELL_IN;
          }
          else
          {
            ReturnCode = CELL_OUTX;      // Change return code later on.
          }
      #if (NB_DIM == 3)
        }
      #endif
    }
  }

  return(ReturnCode);
}

unsigned char GetCell(double pos[], unsigned int NumCell[])
{
  unsigned char ReturnCode;

  #if PERIODIC_CALCUL == 1
    ReturnCode = GetCellPeriod(pos, NumCell);
  #else
    ReturnCode = GetCellNonPeriod(pos, NumCell);
  #endif

  return(ReturnCode); 
}

/* -------------- Calculate number of particles in each cells of the grid ------------- */
void DensityCalculate()
{
  unsigned int i;
  unsigned CR;                                                  
  unsigned int NumCell[3];

  NumCell[2] = 0;
  NbPartOut = 0;
  memset(CellNbPart, 0, sizeof(CellNbPart));
  memset(CellNbPartStetGas, 0, sizeof(CellNbPartStetGas));

  for(i = 0; i < NbPart; i++)
  {
    CR = GetCell(Part[i].pos, NumCell);
    switch(CR)
    {
      case CELL_OUTX:
      case CELL_OUTY:
      case CELL_OUTZ:
        NbPartOut++;
        break;
      case CELL_IN:
	      if(NumCell[0] < 0) Error(100);
	      if(NumCell[1] < 0) Error(101);
	      if(NumCell[2] < 0) Error(102);
	      if(NumCell[0] >= NBCELLX_GRID) Error(103);
	      if(NumCell[1] >= NBCELLY_GRID) Error(104);
	      if(NumCell[2] >= NBCELLZ_GRID) Error(105);
        CellNbPart[NumCell[0]][NumCell[1]][NumCell[2]]++;
        if(i < NbPartStetGas) CellNbPartStetGas[NumCell[0]][NumCell[1]][NumCell[2]]++;
        else ; // NOP
        break;
      default:
        Error(60);
    }
	}
}

/* ----- Calculate next position of a particle which is far from galactic center ------- */
void ProcessPositOut(int PartInd, unsigned int NumCell[])      
{
  double dt2;

  /********* Calculate acceleration from the galaxy ******************/
  GetAccelFromGalaxy(PartInd, Part[PartInd].acc, NumCell);
      
  /**** Calculate next position from previously calculated accel *****/
  dt2 = Gravit_dt * Gravit_dt;  
  Part[PartInd].posnew[0] =  - Part[PartInd].posold[0] + 2 * Part[PartInd].pos[0]
      			                                        + Part[PartInd].acc[0] * dt2;
  Part[PartInd].posnew[1] =  - Part[PartInd].posold[1] + 2 * Part[PartInd].pos[1]
      			                                        + Part[PartInd].acc[1] * dt2;
  #if (NB_DIM == 3)
    Part[PartInd].posnew[2] =  - Part[PartInd].posold[2] + 2 * Part[PartInd].pos[2]
      			                                        + Part[PartInd].acc[2] * dt2;
  #endif
}

/* ----- Calculate next posit of a partic inside the grid ------------- */
void ProcessPositIn(int PartInd, unsigned int NumCell[])      
{
  double dt2, StetCorr;

  /**** Calculate acceleration from previously calculated potential *****/
  #if (NB_DIM == 3)
    if(     NumCell[0]<=0  ||  NumCell[1]<=0  ||  NumCell[2]<=0
        ||  NumCell[0]>=NBCELLX_GRID-1  ||  NumCell[1]>=NBCELLY_GRID-1 
                                        ||  NumCell[2]>=NBCELLZ_GRID-1 )
  #else
    if( NumCell[0]<=0  ||  NumCell[1]<=0  ||  NumCell[0]>=NBCELLX_GRID-1
                                          ||  NumCell[1]>=NBCELLY_GRID-1 )
  #endif
  {    // On the fringe. Impossible to calculate acceleration in a coherent way.
    Part[PartInd].acc[0] = Part[PartInd].acc[1] = 0; // Not perfect, should be avoided.
    #if (NB_DIM == 3)
      Part[PartInd].acc[2] = 0;
    #endif
  }
  else
  {                                /* Done in a "time reversal" manner */
    Part[PartInd].acc[0] =  Potential[NumCell[0]+1][NumCell[1]][NumCell[2]];
    Part[PartInd].acc[0] -= Potential[NumCell[0]-1][NumCell[1]][NumCell[2]];
    Part[PartInd].acc[0] /= 2 * LGX_CELL;
    Part[PartInd].acc[1] =  Potential[NumCell[0]][NumCell[1]+1][NumCell[2]];
    Part[PartInd].acc[1] -= Potential[NumCell[0]][NumCell[1]-1][NumCell[2]];
    Part[PartInd].acc[1] /= 2 * LGY_CELL;
    #if (NB_DIM == 3)
      Part[PartInd].acc[2] =  Potential[NumCell[0]][NumCell[1]][NumCell[2]+1];
      Part[PartInd].acc[2] -= Potential[NumCell[0]][NumCell[1]][NumCell[2]-1];
      Part[PartInd].acc[2] /= 2 * LGZ_CELL;
    #endif

    #if GRAVIT_MODEL == STET  &&  STET_MODE == STET_COMPLEX 
      StetCorr = StetCorrMemo[NumCell[0]][NumCell[1]][NumCell[2]];
      Part[PartInd].acc[0] *= StetCorr;
      Part[PartInd].acc[1] *= StetCorr;
      #if (NB_DIM == 3)
        Part[PartInd].acc[2] *= StetCorr;
      #endif
    #endif
  }

  /**** Calculate next position from previously calculated accel *************/
  dt2 = Gravit_dt * Gravit_dt;  
  Part[PartInd].posnew[0] = - Part[PartInd].posold[0] + 2 * Part[PartInd].pos[0]
      			                                      + Part[PartInd].acc[0] * dt2;
  Part[PartInd].posnew[1] = - Part[PartInd].posold[1] + 2 * Part[PartInd].pos[1]
      			                                      + Part[PartInd].acc[1] * dt2;
  #if (NB_DIM == 3)
    Part[PartInd].posnew[2] = - Part[PartInd].posold[2] + 2 * Part[PartInd].pos[2]
      			                                         + Part[PartInd].acc[2] * dt2;
  #endif
}

/* ------------------ Calculate next position of a particle ----------------- */
void ProcessPosit()      
{
  unsigned int i;
  unsigned CR;
  unsigned int NumCell[3];

  for(i = 0; i < NbPart; i++)
  {
    CR = GetCell(Part[i].pos, NumCell);
    switch(CR)
    {
      case CELL_OUTX:
      case CELL_OUTY:
      case CELL_OUTZ:
        VectorCopy3D(Part[i].NumCell, NumCell);  // For Stet complex only. Used
        ProcessPositOut(i, NumCell);           // for calc stetcorr of PositOut.
        break;
      case CELL_IN:
        ProcessPositIn(i, NumCell);                  
        VectorCopy3D(NumCell, Part[i].NumCell);         // For Stet complex only.
        break;
      default:
        Error(61);
    }
  }                 // End of for loop over the i particles.

  /* ----- Now that new positionS are calculated, update old and news ---------- */
  for(i = 0; i < NbPart; i++)
  {
    VectorCopy(Part[i].pos, Part[i].posold);
    VectorCopy(Part[i].posnew, Part[i].pos);
  }
}

/* -------------- Calculate acceleration coming the whole galaxy -------------- */
void GetAccelFromGalaxy(int PartInd, double accel[], unsigned int NumCell[])
{
  long double SquareDistance;
  double dx0, dx1, dx;
  #if (NB_DIM == 3)
    double dx2;
  #endif
  double acc;                      // Temporary variable memorizing acceleration.
  double GalaxyMass;

  /* ----- get distance^2 from galactic center ----- */
  dx0 = PosOrigin[0] - Part[PartInd].pos[0];
  dx1 = PosOrigin[1] - Part[PartInd].pos[1];
  SquareDistance = dx0*dx0 + dx1*dx1;
  #if (NB_DIM == 3)
    dx2 = PosOrigin[2] - Part[PartInd].pos[2];
    SquareDistance += dx2*dx2;
  #endif
  dx = sqrt(SquareDistance);

  /* ----- Retrieve cell mass ------------------ */
  GalaxyMass = (NbPart - NbPartOut) * PartMass;

  /* ------- force = G * p1_mass * pd2->mass / SquareDistance ---- */
  if (SquareDistance != 0) acc = GalaxyMass * G_CONST / SquareDistance;
  else acc = 0;                     // Rare case: avoiding any effect.
  accel[0] = (dx0/dx) * acc;        // x composant of accel vector.
  accel[1] = (dx1/dx) * acc;        // y composant of accel vector.
  #if (NB_DIM == 3)
    accel[2] = (dx2/dx) * acc;      // z composant of accel vector.
  #endif

  GetAccelFromGalaxyStet(PartInd, accel, NumCell);
}

/* ---------------- Repositionning the grid "around" the galaxy ----------------- */
void RecenterGrid()
{
  unsigned int i;
  double PosNorm, Origin[3];
  
  VectorZero(Origin);
  for(i = 0; i < NbPart; i++)
  {
    VectorAdd(Part[i].pos, Origin, Origin);   // PosOrigin += pos
  }                                            
  VectorDiv(Origin, NbPart, Origin);          // PosOrigin = barycentric of Particles posit.
  VectorNorm(Origin, &PosNorm);               // PosNorm = ||PosOrigin||.

         //  From time to time ? : if(PosNorm >= MAX_GALACTIC_DISPL  &&  MODE_DEBUG != DEBUG_SOLSYST)
  if(MODE_DEBUG != DEBUG_SOLSYST)             // Do this allways if not in test.
  {
    /* ------ RECENTERING grid ----------------------------------------- */
    VectorCopy(Origin, PosOrigin);            // PosOrigin = Origin. RECENTERING grid.

    /* ------ From times to times, RECENTERING plot -------------------- */
    if(PosNorm >= MAX_GALACTIC_DISPL)         // Only from times to times.
    {                                         // Recenter plot.
      unsigned int j;
      for(j = 0; j < NbPart; j++)
      {
        VectorSub(Part[j].pos, PosOrigin, Part[j].pos);    // PartPos = PartPos - PosOrigin.
      }
    }
    else ;                            // Only from time to time in order to watch motion.
  }
  else ;                                      // Let show the galactic motion a little.
}

void GravitFieldCalculate()    // This function is only used at init time for some cases.
{                              // No need for periodic mode.
  unsigned int PartInd;
  unsigned CR;
  unsigned int NumCell[3];
  #if (GRAVIT_MODEL == STET)
    double StetCorr;        // Corrective term added to Newton's acceleration.
    StetCorr = GetStetCorr(StetNbPartInit,0,0,0);
  #endif

  for(PartInd = 0 ; PartInd < NbPart ; PartInd++)
  {
    CR = GetCell(Part[PartInd].pos, NumCell);
    if(CR != CELL_IN) Error(309);
    
    /**** Calculate acceleration from previously calculated potential *****/
    #if (NB_DIM == 3)
      if(     NumCell[0]<=0  ||  NumCell[0]>=NBCELLX_GRID-1
          ||  NumCell[1]<=0  ||  NumCell[1]>=NBCELLY_GRID-1
          ||  NumCell[2]<=0  ||  NumCell[2]>=NBCELLZ_GRID-1
        )
    #else
      if(     NumCell[0]<=0  ||  NumCell[0]>=NBCELLX_GRID-1
          ||  NumCell[1]<=0  ||  NumCell[1]>=NBCELLY_GRID-1
        )
    #endif
    {    // On the fringe. Impossible to calculate acceleration in a coherent way.
      Part[PartInd].acc[0] = Part[PartInd].acc[1] = 0;
      #if (NB_DIM == 3)
        Part[PartInd].acc[2] = 0;
      #endif
    }
    else
    {                                /* Done in a "time reversal" manner */
        Part[PartInd].acc[0] =  Potential[NumCell[0]+1][NumCell[1]][NumCell[2]];
        Part[PartInd].acc[0] -= Potential[NumCell[0]-1][NumCell[1]][NumCell[2]];
        Part[PartInd].acc[0] /= 2 * LGX_CELL;
        Part[PartInd].acc[1] =  Potential[NumCell[0]][NumCell[1]+1][NumCell[2]];
        Part[PartInd].acc[1] -= Potential[NumCell[0]][NumCell[1]-1][NumCell[2]];
        Part[PartInd].acc[1] /= 2 * LGY_CELL;
      #if (NB_DIM == 3)
        Part[PartInd].acc[2] =  Potential[NumCell[0]][NumCell[1]][NumCell[2]+1];
        Part[PartInd].acc[2] -= Potential[NumCell[0]][NumCell[1]][NumCell[2]-1];
        Part[PartInd].acc[2] /= 2 * LGZ_CELL;
      #endif

      #if (GRAVIT_MODEL == STET)
        Part[PartInd].acc[0] *= StetCorr;
        Part[PartInd].acc[1] *= StetCorr;
        #if (NB_DIM == 3)
          Part[PartInd].acc[2] *= StetCorr;
        #endif
      #endif
    }
  }                 // End of for loop over the i particles.
}

#else              // INIT_TYPE != INIT_BULLET_CLUSTER
  void DensityCalculate(){}
  void GravitFieldCalculate(){}
  void RecenterGrid(){}
  unsigned char GetCell(double pos[], unsigned int NumCell[]){return(0);}
  void ProcessPosit(){}      
  particle_t Part[1];   // For a correct generation.
#endif              // INIT_TYPE != INIT_BULLET_CLUSTER
